__author__ = "Mohamed-Rafik Bouguelia"
__license__ = "MIT"
__email__ = "mohamed-rafik.bouguelia@hh.se"

from .individual_anomaly.individual_anomaly_inductive import IndividualAnomalyInductive
from .individual_anomaly.individual_anomaly_transductive import IndividualAnomalyTransductive
from .group_anomaly.group_anomaly import GroupAnomaly
from ._version import __version__

